/**************************************************************************
*  This file is part of the TAL project (Tiny Abstraction Layer)
*
*  Copyright (c) 2013-2018 by Michael Fischer (www.emb4fun.de).
*  All rights reserved.
*
*  Redistribution and use in source and binary forms, with or without 
*  modification, are permitted provided that the following conditions 
*  are met:
*  
*  1. Redistributions of source code must retain the above copyright 
*     notice, this list of conditions and the following disclaimer.
*  2. Redistributions in binary form must reproduce the above copyright
*     notice, this list of conditions and the following disclaimer in the 
*     documentation and/or other materials provided with the distribution.
*  3. Neither the name of the author nor the names of its contributors may 
*     be used to endorse or promote products derived from this software 
*     without specific prior written permission.
*
*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
*  THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
*  OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
*  AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
*  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF 
*  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
*  SUCH DAMAGE.
*
***************************************************************************
*  History:
*
*  19.05.2013  mifi  First Version
**************************************************************************/
#define __TALMISC_C__

/*=======================================================================*/
/*  Include                                                              */
/*=======================================================================*/
#include <string.h>
#include "tal.h"

/*=======================================================================*/
/*  All Structures and Common Constants                                  */
/*=======================================================================*/

/*=======================================================================*/
/*  Definition of all local Data                                         */
/*=======================================================================*/

#if 0
static const uint8_t xorData[128*4] __attribute__((aligned(0x4))) = 
{
   0x7b, 0xf6, 0xcc, 0x39, 0xc1, 0x6d, 0xc1, 0x41, 0x27, 0xe9, 0x53, 0xa2, 0x4f, 0xef, 0xc8, 0x33, 
   0xcc, 0x37, 0x73, 0x41, 0xd6, 0x33, 0xe9, 0x02, 0xd7, 0x3c, 0x2b, 0x1e, 0x87, 0x28, 0x61, 0x46, 
   0xa3, 0x96, 0xf6, 0x54, 0x14, 0x2b, 0xa3, 0x93, 0xab, 0x9e, 0xf9, 0x15, 0xea, 0xb6, 0x6a, 0x7c, 
   0xc8, 0x4a, 0x7b, 0x59, 0x1c, 0x3f, 0x2d, 0xcc, 0x7a, 0xe3, 0x18, 0x38, 0x80, 0xda, 0x41, 0xfc, 
   0x62, 0xb0, 0x6a, 0xb2, 0x02, 0x7e, 0x5d, 0x05, 0x53, 0xab, 0xe7, 0xbf, 0x9e, 0xbe, 0x1b, 0x54, 
   0x32, 0xc1, 0x68, 0x23, 0xe1, 0xbc, 0x44, 0x58, 0xff, 0x13, 0x4f, 0xb5, 0x16, 0x80, 0x7f, 0x53, 
   0x65, 0x2a, 0xae, 0xc1, 0x89, 0x71, 0xdb, 0x4f, 0x5a, 0xb0, 0xb5, 0xd6, 0x72, 0xc3, 0x8c, 0xe9, 
   0xd2, 0xa9, 0x2c, 0x85, 0x91, 0xb5, 0x1f, 0x55, 0xec, 0xcd, 0x13, 0xd0, 0xbf, 0x94, 0xdf, 0xbb, 
   0xdd, 0x3e, 0x63, 0x17, 0xe6, 0x1b, 0x8d, 0x0d, 0x32, 0x6b, 0x10, 0xc2, 0xc3, 0x45, 0x87, 0x64, 
   0x18, 0x33, 0x31, 0x5b, 0x4f, 0x63, 0x93, 0xd9, 0x5b, 0x99, 0xf2, 0x4c, 0xf9, 0xff, 0xd8, 0x23, 
   0xbc, 0xc4, 0xf8, 0x92, 0x1d, 0x98, 0x4a, 0x04, 0x88, 0xe9, 0x64, 0xaa, 0xb8, 0x7c, 0xf9, 0xd8, 
   0x16, 0x9f, 0x40, 0x20, 0xf4, 0x2a, 0x5d, 0x5f, 0xd5, 0xbb, 0xdb, 0x55, 0x3d, 0xc6, 0x33, 0xc6, 
   0x60, 0x52, 0x07, 0xac, 0x62, 0x16, 0x4e, 0x08, 0x1b, 0xe8, 0xbc, 0xa5, 0x44, 0xac, 0x08, 0xdd, 
   0x14, 0x67, 0x6f, 0xcb, 0x5e, 0x90, 0x7a, 0x21, 0xa5, 0x3b, 0x56, 0x04, 0x90, 0x7e, 0x1f, 0x7a, 
   0x56, 0xd5, 0x87, 0x02, 0xf7, 0xff, 0xf4, 0x93, 0xed, 0x60, 0x3f, 0x54, 0x54, 0xca, 0x63, 0xd3, 
   0xa3, 0xe8, 0xf8, 0xbd, 0xe0, 0x13, 0xfa, 0x43, 0x2b, 0xfe, 0xa3, 0x1f, 0xa9, 0xac, 0xd0, 0xb3, 
   0x65, 0xe7, 0x5f, 0x20, 0xd0, 0x7c, 0xa2, 0x4a, 0xa1, 0xbe, 0x7a, 0x59, 0x5c, 0xdd, 0x4f, 0xda, 
   0x59, 0x9e, 0x39, 0x73, 0xf4, 0xf3, 0x38, 0x00, 0x73, 0x9d, 0x5b, 0xf9, 0xae, 0x14, 0x26, 0xa6, 
   0x9d, 0xbf, 0x09, 0xe6, 0xf2, 0x8e, 0x04, 0xea, 0xfc, 0xcc, 0x78, 0xe7, 0xee, 0xe1, 0x79, 0xd9, 
   0xbc, 0x0b, 0xe8, 0x9f, 0x5a, 0xc6, 0xed, 0xde, 0x85, 0xcb, 0x5e, 0x1d, 0x3c, 0x11, 0x45, 0x78, 
   0x66, 0xa4, 0x5b, 0x2d, 0x8e, 0xf2, 0x03, 0x77, 0x84, 0x85, 0xec, 0x9a, 0x2f, 0x6b, 0x3f, 0x5a, 
   0x04, 0x44, 0xff, 0x84, 0xe3, 0xb8, 0x10, 0x61, 0xec, 0x82, 0xa3, 0xc2, 0xd3, 0x07, 0x0d, 0xef, 
   0x29, 0x5c, 0xc9, 0x95, 0x06, 0xb7, 0xb2, 0xdb, 0x67, 0x12, 0xfb, 0x49, 0x30, 0x72, 0x50, 0x0a, 
   0x8a, 0x0e, 0x79, 0x8f, 0x5e, 0x51, 0x7e, 0x70, 0x8c, 0x07, 0x7a, 0x08, 0xd9, 0xf3, 0xd4, 0xba, 
   0xab, 0x4a, 0x4b, 0xe9, 0x63, 0x94, 0xa3, 0x47, 0x68, 0x6c, 0x45, 0xbe, 0x9c, 0xed, 0x84, 0xa8, 
   0xef, 0x3c, 0x54, 0x7b, 0x3b, 0xdc, 0x25, 0x71, 0x48, 0xe5, 0xfa, 0xb6, 0x56, 0x07, 0x91, 0x31, 
   0x3b, 0xd8, 0x09, 0x89, 0xb2, 0x25, 0x33, 0xd5, 0xab, 0xc5, 0x45, 0x74, 0xf5, 0xfc, 0x29, 0x8f, 
   0x86, 0x61, 0xb7, 0xf4, 0x51, 0x16, 0xf9, 0x7c, 0xda, 0x70, 0x89, 0xa7, 0x94, 0xb9, 0x0d, 0x73, 
   0x35, 0x7e, 0x91, 0xc8, 0x58, 0xa2, 0x00, 0xda, 0x58, 0xa3, 0xe7, 0x3a, 0x5b, 0xe2, 0xa6, 0xa3, 
   0x7d, 0xa1, 0x3f, 0xc0, 0x64, 0x55, 0x7f, 0xcc, 0x3c, 0x1d, 0xad, 0x3c, 0x78, 0xf8, 0x26, 0x98, 
   0x9d, 0x8b, 0x38, 0x47, 0x42, 0x94, 0xfc, 0x26, 0xf8, 0x23, 0x72, 0x5f, 0xa9, 0xed, 0x96, 0xca, 
   0x73, 0x49, 0x88, 0x1e, 0x09, 0x1c, 0x16, 0x90, 0xc2, 0x4d, 0xec, 0x89, 0x9a, 0xd0, 0x1f, 0xc4
};
#endif

/*=======================================================================*/
/*  Definition of all local Procedures                                   */
/*=======================================================================*/

/*=======================================================================*/
/*  All code exported                                                    */
/*=======================================================================*/

/*************************************************************************/
/*  tal_MISCRingSetup                                                    */
/*                                                                       */
/*  Setup the ring buffer.                                               */
/*                                                                       */
/*  In    : pRing, pBuffer, wElementSize, wElementCount                  */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void tal_MISCRingSetup (TAL_MISC_RING *pRing, 
                        uint8_t       *pBuffer, 
                        uint16_t       wElementSize,
                        uint16_t       wElementCount)
{
   pRing->pBuffer          = pBuffer;
   pRing->wElementSize     = wElementSize;
   pRing->wMaxElementCount = wElementCount; 
   pRing->wInIndex         = 0;
   pRing->wOutIndex        = 0;
   pRing->wCount           = 0;    
} /* tal_MISCRingSetup */

/*************************************************************************/
/*  tal_MISCRingReset                                                    */
/*                                                                       */
/*  Reset the ring buffer.                                               */
/*                                                                       */
/*  In    : pRing                                                        */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void tal_MISCRingReset (TAL_MISC_RING *pRing)
{
   TAL_CPU_DISABLE_ALL_INTS();

   pRing->wInIndex  = 0;
   pRing->wOutIndex = 0;
   pRing->wCount    = 0;
   
   TAL_CPU_ENABLE_ALL_INTS();
} /* tal_MISCRingReset */

/*************************************************************************/
/*  tal_MISCRingGetFreeCount                                             */
/*                                                                       */
/*  Return the free count of the ring buffer.                            */
/*                                                                       */
/*  In    : pRing                                                        */
/*  Out   : none                                                         */
/*  Return: Count                                                        */
/*************************************************************************/
uint16_t tal_MISCRingGetFreeCount (TAL_MISC_RING *pRing)
{
   uint16_t wCount;
   
   TAL_CPU_DISABLE_ALL_INTS();

   wCount = pRing->wMaxElementCount - pRing->wCount;
   
   TAL_CPU_ENABLE_ALL_INTS();
   
   return(wCount);
} /* tal_MISCRingGetFreeCount */

/*************************************************************************/
/*  tal_MISCRingGetUseCount                                              */
/*                                                                       */
/*  Return the used count of the ring buffer.                            */
/*                                                                       */
/*  In    : pRing                                                        */
/*  Out   : none                                                         */
/*  Return: Count                                                        */
/*************************************************************************/
uint16_t tal_MISCRingGetUseCount (TAL_MISC_RING *pRing)
{
   uint16_t wCount;
   
   TAL_CPU_DISABLE_ALL_INTS();

   wCount = pRing->wCount;
   
   TAL_CPU_ENABLE_ALL_INTS();
   
   return(wCount);
} /* tal_MISCRingGetUseCount */

/*************************************************************************/
/*  tal_MISCRingAdd                                                      */
/*                                                                       */
/*  Add a new element to the ring buffer.                                */
/*                                                                       */
/*  In    : pRing, pData                                                 */
/*  Out   : none                                                         */
/*  Return: TAL_OK / TAL_ERROR                                           */
/*************************************************************************/
TAL_RESULT tal_MISCRingAdd (TAL_MISC_RING *pRing, uint8_t *pData)
{
   TAL_RESULT Error = TAL_ERR_RING_FULL;
   uint16_t  wIndex;

   TAL_CPU_DISABLE_ALL_INTS();
   
   /* Check if a free entry is available */
   if (pRing->wCount != pRing->wMaxElementCount)
   {
      pRing->wCount++;
      
      wIndex = (uint16_t)(pRing->wInIndex * pRing->wElementSize);
      
      memcpy(&pRing->pBuffer[wIndex], pData, pRing->wElementSize);
      
      pRing->wInIndex++;
      
      if (pRing->wInIndex == pRing->wMaxElementCount)
      {
         pRing->wInIndex = 0;
      }
      
      Error = TAL_OK;
   }

   TAL_CPU_ENABLE_ALL_INTS();
   
   return(Error);
} /* tal_MISCRingAdd */

/*************************************************************************/
/*  tal_MISCRingGet                                                      */
/*                                                                       */
/*  Get an element from the ring buffer.                                 */
/*                                                                       */
/*  In    : pRing, pData                                                 */
/*  Out   : pData                                                        */
/*  Return: TAL_OK / TAL_ERROR                                           */
/*************************************************************************/
TAL_RESULT tal_MISCRingGet (TAL_MISC_RING *pRing, uint8_t *pData)
{
   TAL_RESULT Error = TAL_ERR_RING_EMPTY;
   uint16_t  wIndex;

   TAL_CPU_DISABLE_ALL_INTS();
   
   /* Check if an entry is available */
   if (pRing->wCount != 0)
   {
      pRing->wCount--;
      
      wIndex = (uint16_t)(pRing->wOutIndex * pRing->wElementSize);
      
      memcpy(pData, &pRing->pBuffer[wIndex], pRing->wElementSize);
      
      pRing->wOutIndex++;
      
      if (pRing->wOutIndex == pRing->wMaxElementCount)
      {
         pRing->wOutIndex = 0;
      }
      
      Error = TAL_OK;
   }

   TAL_CPU_ENABLE_ALL_INTS();

   return(Error);   
} /* tal_MISCRingGet */

#if 0
/*************************************************************************/
/*  tal_MISCRandomInit                                                   */
/*                                                                       */
/*  Initialize the random generator.                                     */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void tal_MISCRandomInit (void)
{
} /* tal_MISCRandomInit */

/*************************************************************************/
/*  tal_MISCGetRandomData                                                */
/*                                                                       */
/*  Computes a pseudo-random integers in the range 0 to 2^32.            */
/*                                                                       */
/*  In    : dSeed                                                        */
/*  Out   : none                                                         */
/*  Return: the computed pseudo-random integer.                          */
/*************************************************************************/
uint32_t tal_MISCGetRandomData (uint32_t dSeed)
{
   unsigned int       *xorTable = (unsigned int*)xorData;
   static unsigned int Y = 31468; // your seed
   static unsigned int iterations = 0, lastUpper = 0;
   unsigned int r, n, pos1, pos2, choice;
   
   if (dSeed != 0)
   {
      Y          = dSeed;
      iterations = 0;
      lastUpper  = 0;
   }
   
   // begin critical section
   r = Y;
   n = iterations;
   choice = lastUpper;
   // end critical section
   
   pos1 = (r ^ xorTable[choice & 0x3f]) + xorTable[n & 0x3f];
   pos2 = ~pos1;
   
   r += ~r + ~n ^
      (xorTable[(n+r) & 0x7f]    << (((r & 0x1f000000) >> 24))) ^
      (xorTable[choice & 0x7f]   << (((r & 0x000f0000) >> 16))) ^
      (xorTable[pos2 & 0x7f]     << (((r & 0x00001f00) >> 8)))  ^
      (xorTable[pos1 & 0x7f]     << (((r & 0x0000000f) )))      ^
      (xorTable[(n+pos1) & 0x7f] >> (((r & 0x1f000000) >> 24))) ^
      (xorTable[choice & 0x7f]   >> (((r & 0x001f0000) >> 16))) ^
      (xorTable[pos2 & 0x7f]     >> (((r & 0x00000f00) >> 8)))  ^
      (xorTable[(r+pos2) & 0x7f] >> (((r & 0x0000001f) )));
      
   //choice=r>>24;
   
   // begin critical section
   Y = r;
   iterations++;
   lastUpper = r >> 24;
   // end critical section
   
   return(Y);
} /* tal_MISCGetRandomData */
#endif

/*** EOF ***/
